# -*- coding: utf-8 -*-
#
#  Copyright (C) 2001, 2002 by Tamito KAJIYAMA
#  Copyright (C) 2002, 2003 by MATSUMURA Namihiko <nie@counterghost.net>
#  Copyright (C) 2002-2013 by Shyouzou Sugitani <shy@users.sourceforge.jp>
#  Copyright (C) 2003 by Shun-ichi TAHARA <jado@flowernet.gr.jp>
#
#  This program is free software; you can redistribute it and/or modify it
#  under the terms of the GNU General Public License (version 2) as
#  published by the Free Software Foundation.  It is distributed in the
#  hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
#  PURPOSE.  See the GNU General Public License for more details.
#

import abc

from gi.repository import Gtk
from gi.repository import Gdk
from gi.repository import GdkPixbuf
from gi.repository import Pango
from gi.repository import PangoCairo
from gi.repository import GLib
import cairo

import ninix.pix


class Balloon:

    def __init__(self):
        self.request_parent = lambda *a: None # dummy
        self.synchronized = []
        self.user_interaction = False
        self.window = []
        # create communicatebox
        self.communicatebox = CommunicateBox()
        self.communicatebox.set_responsible(self.handle_request)
        # create teachbox
        self.teachbox = TeachBox()
        self.teachbox.set_responsible(self.handle_request)
        # create inputbox
        self.inputbox = InputBox()
        self.inputbox.set_responsible(self.handle_request)
        # create passwordinputbox
        self.passwordinputbox = PasswordInputBox()
        self.passwordinputbox.set_responsible(self.handle_request)

    def set_responsible(self, request_method):
        self.request_parent = request_method

    def handle_request(self, event_type, event, *arglist, **argdict):
        assert event_type in ['GET', 'NOTIFY']
        handlers = {
            'reset_user_interaction': self.reset_user_interaction,
            }
        handler = handlers.get(event)
        if handler is None:
            result = self.request_parent(
                event_type, event, *arglist, **argdict)
        else:
            result = handler(*arglist, **argdict)
        if event_type == 'GET':
            return result

    def reset_user_interaction(self):
        self.user_interaction = False

    def get_text_count(self, side):
        if len(self.window) > side:
            return self.window[side].get_text_count()
        else:
            return 0

    def get_window(self, side):
        if len(self.window) > side:
            return self.window[side].window ## FIXME
        else:
            return None

    def reset_text_count(self, side):
        if len(self.window) > side:
            self.window[side].reset_text_count()

    def reset_balloon(self):
        for balloon_window in self.window:
            balloon_window.reset_balloon()

    def create_gtk_window(self, title):
        window = ninix.pix.TransparentWindow()
        window.set_focus_on_map(False)
        window.set_title(title)
        window.set_decorated(False)
        window.set_resizable(False)
        window.set_skip_pager_hint(False)
        window.set_skip_taskbar_hint(True)
        window.connect('delete_event', self.delete)
        window.realize()
        return window

    def identify_window(self, win):
        for balloon_window in self.window:
            if win == balloon_window.window.get_window():
                return True
        return False

    def delete(self, window, event):
        return True

    def finalize(self):
        for balloon_window in self.window:
            balloon_window.destroy()
        self.window = []
        self.communicatebox.destroy()
        self.teachbox.destroy()
        self.inputbox.destroy()
        self.passwordinputbox.destroy()

    def new(self, desc, balloon):
        self.desc = desc
        self.directory = balloon['balloon_dir'][0]
        balloon0 = {}
        balloon1 = {}
        communicate0 = None
        communicate1 = None
        communicate2 = None
        communicate3 = None
        for key, value in balloon.items():
            if key in ['arrow0', 'arrow1']:
                balloon0[key] = value
                balloon1[key] = value
            elif key == 'sstp':
                balloon0[key] = value  # sstp marker
            elif key.startswith('s'):
                balloon0[key] = value  # Sakura
            elif key.startswith('k'):
                balloon1[key] = value  # Unyuu
            elif key == 'c0':
                communicate0 = value # send box
            elif key == 'c1':
                communicate1 = value # communicate box
            elif key == 'c2':
                communicate2 = value # teach box
            elif key == 'c3':
                communicate3 = value # input box
        self.balloon0 = balloon0
        self.balloon1 = balloon1
        # create balloon windows
        for balloon_window in self.window:
            balloon_window.destroy()
        self.window = []
        self.add_window(0)
        self.add_window(1)
        # configure communicatebox
        self.communicatebox.new(desc, communicate1)
        # configure teachbox
        self.teachbox.new(desc, communicate2)
        # configure inputbox
        self.inputbox.new(desc, communicate3)
        # configure passwordinputbox
        self.passwordinputbox.new(desc, communicate3)

    def add_window(self, side):
        assert len(self.window) == side
        if side == 0:
            name = 'balloon.sakura'
            id_format = 's{0:d}'
            balloon = self.balloon0
        elif side == 1:
            name = 'balloon.kero'
            id_format = 'k{0:d}'
            balloon = self.balloon1
        else:
            name = 'balloon.char{0:d}'.format(side)
            id_format = 'k{0:d}'
            balloon = self.balloon1
        gtk_window = self.create_gtk_window(name)
        balloon_window = BalloonWindow(
            gtk_window, side, self.desc, balloon,
            id_format)
        balloon_window.set_responsible(self.handle_request)
        self.window.append(balloon_window)

    def reset_fonts(self):
        for window in self.window:
            window.reset_fonts()

    def get_balloon_directory(self):
        return self.directory

    def get_balloon_size(self, side):
        if len(self.window) > side:
            return self.window[side].get_balloon_size()
        else:
            return (0, 0)

    def get_balloon_windowposition(self, side):
        if len(self.window) > side:
            return self.window[side].get_balloon_windowposition()
        else:
            return (0, 0)

    def set_balloon_default(self):
        default_id = self.request_parent('GET', 'get_balloon_default_id')
        try:
            default_id = int(default_id)
        except:
            default_id = 0
        for side in range(len(self.window)):
            self.window[side].set_balloon(default_id)

    def set_balloon(self, side, num):
        if len(self.window) > side:
            self.window[side].set_balloon(num)

    def set_position(self, side, base_x, base_y):
        if len(self.window) > side:
            self.window[side].set_position(base_x, base_y)

    def get_position(self, side):
        if len(self.window) > side:
            return self.window[side].get_position()
        else:
            return (0, 0)

    def set_autoscroll(self, flag):
        for side in range(len(self.window)):
            self.window[side].set_autoscroll(flag)

    def is_shown(self, side):
        return 0 if len(self.window) <= side else self.window[side].is_shown()

    def show(self, side):
        if len(self.window) > side:
            self.window[side].show()

    def hide_all(self):
        for side in range(len(self.window)):
            self.window[side].hide()

    def hide(self, side):
        if len(self.window) > side:
            self.window[side].hide()

    def raise_all(self):
        for side in range(len(self.window)):
            self.window[side].raise_()

    def raise_(self, side):
        if len(self.window) > side:
            self.window[side].raise_()

    def lower_all(self):
        for side in range(len(self.window)):
            self.window[side].lower()

    def lower(self, side):
        if len(self.window) > side:
            self.window[side].lower()

    def synchronize(self, list):
        self.synchronized = list

    def clear_text_all(self):
        for side in range(len(self.window)):
            self.clear_text(side)

    def clear_text(self, side):
        if self.synchronized:
            for side in self.synchronized:
                if len(self.window) > side:
                    self.window[side].clear_text()
        else:
            if len(self.window) > side:
                self.window[side].clear_text()

    def append_text(self, side, text):
        if self.synchronized:
            for side in self.synchronized:
                if len(self.window) > side:
                    self.window[side].append_text(text)
        else:
            if len(self.window) > side:
                self.window[side].append_text(text)

    def append_sstp_marker(self, side):
        if len(self.window) > side:
            self.window[side].append_sstp_marker()

    def append_link_in(self, side, label):
        if self.synchronized:
            for side in self.synchronized:
                if len(self.window) > side:
                    self.window[side].append_link_in(label)
        else:
            if len(self.window) > side:
                self.window[side].append_link_in(label)

    def append_link_out(self, side, label, value):
        if self.synchronized:
            for side in self.synchronized:
                if len(self.window) > side:
                    self.window[side].append_link_out(label, value)
        else:
            if len(self.window) > side:
                self.window[side].append_link_out(label, value)

    def append_link(self, side, label, value, newline_required=0):
        if self.synchronized:
            for side in self.synchronized:
                if len(self.window) > side:
                    self.window[side].append_link_in(label)
                    self.window[side].append_text(value)
                    self.window[side].append_link_out(label, value)
                    if newline_required:
                        self.window[side].set_newline()
        else:
            if len(self.window) > side:
                self.window[side].append_link_in(label)
                self.window[side].append_text(value)
                self.window[side].append_link_out(label, value)
                if newline_required:
                    self.window[side].set_newline()

    def append_meta(self, side, tag):
        if self.synchronized:
            for side in self.synchronized:
                if len(self.window) > side:
                    self.window[side].append_meta(tag)
        else:
            if len(self.window) > side:
                self.window[side].append_meta(tag)

    def append_image(self, side, path, x, y):
        if len(self.window) > side:
            self.window[side].append_image(path, x, y)

    def show_sstp_message(self, message, sender):
        self.window[0].show_sstp_message(message, sender)

    def hide_sstp_message(self):
        self.window[0].hide_sstp_message()

    def open_communicatebox(self):
        if not self.user_interaction:
            self.user_interaction = True
            self.communicatebox.show()

    def open_teachbox(self):
        if not self.user_interaction:
            self.user_interaction = True
            self.request_parent('NOTIFY', 'notify_event', 'OnTeachStart')
            self.teachbox.show()

    def open_inputbox(self, symbol, limittime=-1, default=None):
        if not self.user_interaction:
            self.user_interaction = True
            self.inputbox.set_symbol(symbol)
            self.inputbox.set_limittime(limittime)
            self.inputbox.show(default)

    def open_passwordinputbox(self, symbol, limittime=-1, default=None):
        if not self.user_interaction:
            self.user_interaction = True
            self.passwordinputbox.set_symbol(symbol)
            self.passwordinputbox.set_limittime(limittime)
            self.passwordinputbox.show(default)

    def close_inputbox(self, symbol):
        if not self.user_interaction:
            return
        self.inputbox.close(symbol)
        self.passwordinputbox.close(symbol)


class BalloonWindow:

    def __init__(self, window, side, desc, balloon, id_format):
        self.window = window
        self.side = side
        self.request_parent = lambda *a: None # dummy
        self.desc = desc
        self.balloon = balloon
        self.balloon_id = None
        self.id_format = id_format
        self.num = 0
        self.__shown = 0
        self.sstp_marker = []
        self.sstp_region = None
        self.sstp_message = None
        self.images = []
        self.width = 0
        self.height = 0
        self.__font_name = ''
        self.text_count = 0
        self.balloon_surface = None
        self.autoscroll = True
        self.dragged = False
        self.x_root = None
        self.y_root = None
        self.x_fractions = 0
        self.y_fractions = 0
        # create drawing area
        self.darea = Gtk.DrawingArea()
        self.darea.set_events(Gdk.EventMask.EXPOSURE_MASK|
                              Gdk.EventMask.BUTTON_PRESS_MASK|
                              Gdk.EventMask.BUTTON_RELEASE_MASK|
                              Gdk.EventMask.POINTER_MOTION_MASK|
                              Gdk.EventMask.POINTER_MOTION_HINT_MASK|
                              Gdk.EventMask.SCROLL_MASK)
        self.darea.connect('draw', self.redraw)
        self.darea.connect('button_press_event', self.button_press)
        self.darea.connect('button_release_event', self.button_release)
        self.darea.connect('motion_notify_event', self.motion_notify)
        self.darea.connect('scroll_event', self.scroll)
        self.window.add(self.darea)
        self.window.realize()
        self.window.override_background_color(Gtk.StateFlags.NORMAL,
                                              Gdk.RGBA(0, 0, 0, 0))
        self.layout = Pango.Layout(self.darea.get_pango_context())
        self.sstp_layout = Pango.Layout(self.darea.get_pango_context())
        mask_r = desc.get_with_type('maskcolor.r', int, 128)
        mask_g = desc.get_with_type('maskcolor.g', int, 128)
        mask_b = desc.get_with_type('maskcolor.b', int, 128)
        self.cursor_color = (mask_r / 255., mask_g / 255., mask_b / 255.)
        text_r = desc.get_with_type(['font.color.r', 'fontcolor.r'], int, 0)
        text_g = desc.get_with_type(['font.color.g', 'fontcolor.g'], int, 0)
        text_b = desc.get_with_type(['font.color.b', 'fontcolor.b'], int, 0)
        self.text_normal_color = (text_r / 255., text_g / 255., text_b / 255.)
        if desc.get_with_type('maskmethod', int) == 1:
            text_r = 255 - text_r
            text_g = 255 - text_g
            text_b = 255 - text_b
        self.text_active_color = (text_r / 255., text_g / 255., text_b / 255.)
        sstp_r = desc.get_with_type('sstpmessage.font.color.r', int, text_r)
        sstp_g = desc.get_with_type('sstpmessage.font.color.g', int, text_g)
        sstp_b = desc.get_with_type('sstpmessage.font.color.b', int, text_b)
        self.sstp_message_color = (sstp_r / 255., sstp_g / 255., sstp_b / 255.)
        # initialize
        self.__direction = min(side, 1) ## kluge: multi character
        self.position = (0, 0)
        self.reset_fonts()
        self.clear_text()

    def set_responsible(self, request_method):
        self.request_parent = request_method

    @property
    def scale(self):
        scaling = self.request_parent('GET', 'get_preference', 'balloon_scaling')
        scale = self.request_parent('GET', 'get_preference', 'surface_scale')
        return scale if scaling else 100 # [%]

    @property
    def direction(self):
        return self.__direction

    @direction.setter
    def direction(self, direction):
        if self.__direction != direction:
            self.__direction = direction # 0: left, 1: right
            self.reset_balloon()

    def get_balloon_windowposition(self):
        winpos_x = self.__get_with_scaling('windowposition.x', int, 0)
        winpos_y = self.__get_with_scaling('windowposition.y', int, 0)
        return winpos_x, winpos_y

    def get_image_surface(self, balloon_id):
        if balloon_id not in self.balloon:
            return None
        try:
            path, config = self.balloon[balloon_id]
            use_pna = self.request_parent('GET', 'get_preference', 'use_pna')
            surface = ninix.pix.create_surface_from_file(
                path, use_pna=use_pna)
        except:
            return None
        return surface

    def reset_fonts(self):
        font_name = self.request_parent('GET', 'get_preference', 'balloon_fonts')
        if self.__font_name == font_name:
            return
        self.font_desc = Pango.FontDescription(font_name)
        pango_size = self.font_desc.get_size()
        if pango_size == 0:
            default_size = 12 # for Windows environment
            size = self.desc.get_with_type(
                ['font.height', 'font.size'], int, default_size)
            pango_size = size * 3 / 4 # convert from Windows to GTK+
            pango_size *= Pango.SCALE
        self.font_desc.set_size(pango_size)
        self.__font_name = font_name
        self.layout.set_font_description(self.font_desc)
        self.layout.set_wrap(Pango.WrapMode.CHAR) # XXX
        # font for sstp message
        if self.side == 0:
            self.sstp_font_desc = Pango.FontDescription(font_name)
            pango_size = self.sstp_font_desc.get_size()
            if pango_size == 0:
                default_size = 10 # for Windows environment
                size = self.desc.get_with_type(
                    'sstpmessage.font.height', int, default_size)
                pango_size = size * 3 / 4 # convert from Windows to GTK+
                pango_size *= Pango.SCALE
            self.sstp_font_desc.set_size(pango_size)
            self.sstp_layout.set_font_description(self.sstp_font_desc)
            self.sstp_layout.set_wrap(Pango.WrapMode.CHAR)
        if self.balloon_id is not None:
            self.reset_message_regions()
            if self.__shown:
                self.darea.queue_draw()

    def reset_sstp_marker(self):
        if self.side == 0:
            assert self.balloon_surface is not None
            w = self.balloon_surface.get_width()
            h = self.balloon_surface.get_height()
            # sstp marker position
            self.sstp = []
            x = self.config_adjust('sstpmarker.x', w,  30)
            y = self.config_adjust('sstpmarker.y', h, -20)
            self.sstp.append((x, y)) # sstp marker
            x = self.config_adjust('sstpmessage.x', w,  50)
            y = self.config_adjust('sstpmessage.y', h, -20)
            self.sstp.append((x, y)) # sstp message
        # sstp marker surface (not only for self.side == 0)
        self.sstp_surface = self.get_image_surface('sstp')

    def reset_arrow(self):
        # arrow positions
        self.arrow = []
        assert self.balloon_surface is not None
        w = self.balloon_surface.get_width()
        h = self.balloon_surface.get_height()
        x = self.config_adjust('arrow0.x', w, -10)
        y = self.config_adjust('arrow0.y', h,  10)
        self.arrow.append((x, y))
        x = self.config_adjust('arrow1.x', w, -10)
        y = self.config_adjust('arrow1.y', h, -20)
        self.arrow.append((x, y))
        # arrow surfaces and sizes
        self.arrow0_surface = self.get_image_surface('arrow0')
        self.arrow1_surface = self.get_image_surface('arrow1')

    def reset_message_regions(self):
        w, h = self.layout.get_pixel_size()
        self.font_height = h
        self.line_space = 1
        self.layout.set_spacing(self.line_space)
        # font metrics
        origin_x = self.__get_with_type(
            'origin.x', int,
            self.__get_with_type(
                'zeropoint.x', int,
                self.__get_with_type('validrect.left', int, 14)))
        origin_y = self.__get_with_type(
            'origin.y', int,
            self.__get_with_type(
                'zeropoint.y', int,
                self.__get_with_type('validrect.top', int, 14)))
        wpx = self.__get_with_type(
            'wordwrappoint.x', int, 
            self.__get_with_type('validrect.right', int, -14))
        if wpx > 0:
            line_width = wpx - origin_x
        elif wpx < 0:
            line_width = self.width - origin_x + wpx
        else:
            line_width = self.width - origin_x * 2
        wpy = self.__get_with_type('validrect.bottom', int, -14)
        if wpy > 0:
            text_height = min(wpy, self.height) - origin_y
        elif wpy < 0:
            text_height = self.height - origin_y + wpy
        else:
            text_height = self.height - origin_y * 2
        line_height = self.font_height + self.line_space
        self.lines = text_height // line_height
        self.line_regions = []
        y = origin_y
        for _ in range(self.lines + 1):
            self.line_regions.append((origin_x, y, line_width, line_height))
            y = y + line_height
        self.line_width = line_width
        # sstp message region
        if self.side == 0:
            w, h = self.sstp_layout.get_pixel_size()
            x, y = self.sstp[1]
            w = line_width + origin_x - x
            self.sstp_region = (x, y, w, h)

    def update_line_regions(self, offset, new_y):
        origin_y = self.__get_with_type(
            'origin.y', int, 
            self.__get_with_type(
                'zeropoint.y', int,
                self.__get_with_type('validrect.top', int, 14)))
        wpy = self.__get_with_type('validrect.bottom', int, -14)
        if wpy > 0:
            text_height = min(wpy, self.height) - origin_y
        elif wpy < 0:
            text_height = self.height - origin_y + wpy
        else:
            text_height = self.height - origin_y * 2
        line_height = self.font_height + self.line_space
        origin_x, y, line_width, line_height = self.line_regions[offset]
        self.lines = offset + (text_height - new_y) // line_height
        y = new_y
        for i in range(offset, len(self.line_regions)):
            self.line_regions[i] = (origin_x, y, line_width, line_height)
            y += line_height
        for i in range(len(self.line_regions), self.lines + 1):
            self.line_regions.append((origin_x, y, line_width, line_height))
            y += line_height

    def get_balloon_size(self, scaling=True):
        w = self.width
        h = self.height
        scale = self.scale
        if scaling:
            w = int(w * scale / 100.0)
            h = int(h * scale / 100.0)
        return w, h

    def reset_balloon(self):
        self.set_balloon(self.num)

    def set_balloon(self, num):
        self.num = num
        balloon_id = self.id_format.format(int(num * 2 + self.direction))
        self.balloon_surface = self.get_image_surface(balloon_id)
        if self.balloon_surface is None:
            balloon_id = self.id_format.format(int(0 + self.direction))
            self.balloon_surface = self.get_image_surface(balloon_id)
        assert self.balloon_surface is not None
        self.balloon_id = balloon_id
        # change surface and window position
        x, y = self.position
        self.width = self.balloon_surface.get_width()
        self.height = self.balloon_surface.get_height()
        scale = self.scale
        w = int(self.width * scale / 100.0)
        h = int(self.height * scale / 100.0)
        self.darea.set_size_request(w, h)
        self.reset_arrow()
        self.reset_sstp_marker()
        self.reset_message_regions()
        self.request_parent('NOTIFY', 'position_balloons')
        if self.__shown:
            self.darea.queue_draw()

    def set_autoscroll(self, flag):
        self.autoscroll = bool(flag)

    def config_adjust(self, name, base, default_value):
        path, config = self.balloon[self.balloon_id]
        value = config.get_with_type(name, int)
        if value is None:
            value = self.desc.get_with_type(name, int)
        if value is None:
            value = default_value
        if value < 0:
            value = base + value
        return int(value)

    def __get_with_type(self, name, conv, default_value):
        path, config = self.balloon[self.balloon_id]
        value = config.get_with_type(name, conv)
        if value is None:
            value = self.desc.get_with_type(name, conv)
            if value is None:
                value = default_value
        return conv(value)

    def __get_with_scaling(self, name, conv, default_value):
        path, config = self.balloon[self.balloon_id]
        value = config.get_with_type(name, conv)
        if value is None:
            value = self.desc.get_with_type(name, conv)
            if value is None:
                value = default_value
        return conv(value * self.scale / 100)

    def __move(self):
        x, y = self.get_position()
        self.window.resize_move(x, y)

    def set_position(self, base_x, base_y):
        if self.balloon_id is None: ## FIXME
            return
        px, py = self.get_balloon_windowposition()
        w, h = self.get_balloon_size()
        if self.direction == 0:
            x = base_x + px - w
        else:
            x = base_x + px
        y = base_y + py
        left, top, scrn_w, scrn_h = ninix.pix.get_workarea()
        if y + h > scrn_h: # XXX
            y = scrn_h - h
        if y < top: # XXX
            y = top
        self.position = (x, y)
        self.__move()

    def get_position(self):
        return self.position

    def destroy(self, finalize=0):
        self.window.remove(self.darea)
        self.darea.destroy()
        self.window.destroy()

    def is_shown(self):
        return 1 if self.__shown else 0

    def show(self):
        if self.__shown:
            return
        self.__shown = 1
        # make sure window is in its position (call before showing the window)
        self.__move()
        self.darea.show()
        self.window.show()
        # make sure window is in its position (call after showing the window)
        self.__move()
        self.raise_()

    def hide(self):
        if not self.__shown:
            return
        self.window.hide()
        self.__shown = 0
        self.images = []

    def raise_(self):
        if self.__shown:
            ##self.window.get_window().raise_()
            getattr(self.window.get_window(), 'raise')() # XXX

    def lower(self):
        if self.__shown:
            self.window.get_window().lower()

    def show_sstp_message(self, message, sender):
        if self.sstp_region is None:
            self.show()
        self.sstp_message = '{0} ({1})'.format(message, sender)
        x, y, w, h = self.sstp_region
        self.sstp_layout.set_text(self.sstp_message, -1)
        message_width, message_height = self.sstp_layout.get_pixel_size()
        if message_width > w:
            self.sstp_message = '... ({0})'.format(sender)
            i = 0
            while 1:
                i += 1
                s = '{0}... ({1})'.format(message[:i], sender)
                self.sstp_layout.set_text(s, -1)
                message_width, message_height = \
                               self.sstp_layout.get_pixel_size()
                if message_width > w:
                    break
                self.sstp_message = s
        self.darea.queue_draw()

    def hide_sstp_message(self):
        self.sstp_message = None
        self.darea.queue_draw()

    def redraw_sstp_message(self, widget, cr):
        if self.sstp_message is None:
            return
        cr.save()
        # draw sstp marker
        if self.sstp_surface is not None:
            x, y = self.sstp[0]            
            cr.set_source_surface(self.sstp_surface, x, y)
            cr.paint()
        # draw sstp message
        x, y, w, h = self.sstp_region
        self.sstp_layout.set_text(self.sstp_message, -1)
        cr.set_source_rgba(*self.sstp_message_color)
        cr.move_to(x, y)
        PangoCairo.update_layout(cr, self.sstp_layout)
        PangoCairo.show_layout(cr, self.sstp_layout)
        cr.restore()

    def redraw_arrow0(self, widget, cr):
        if self.lineno <= 0:
            return
        cr.save()
        x, y = self.arrow[0]
        cr.set_source_surface(self.arrow0_surface, x, y)
        cr.paint()
        cr.restore()

    def redraw_arrow1(self, widget, cr):
        if self.lineno + self.lines >= len(self.text_buffer):
            return
        cr.save()
        x, y = self.arrow[1]
        cr.set_source_surface(self.arrow1_surface, x, y)
        cr.paint()
        cr.restore()

    def set_markup(self, index, text):
        tags_ = ('sup', 'sub', 's', 'u')
        count_ = {}
        for tag_ in tags_:
            count_[tag_] = 0
        markup_list = []
        for sl, sn, tag in self.meta_buffer:
            if sl == index:
                markup_list.append((sn, tag))
        if not markup_list:
            return GLib.markup_escape_text(text)
        markup_list.sort()
        markup_list.reverse()
        pn = len(text)
        for sn, tag in markup_list:
            text = ''.join((text[:sn], tag,
                            GLib.markup_escape_text(text[sn:pn]), text[pn:]))
            pn = sn
            if tag[1] == '/':
                tag_ = tag[2:-1]
                assert tag_ in tags_
                count_[tag_] -= 1
                if count_[tag_] < 0:
                    text = ''.join(('<', tag_, '>', text))
                    count_[tag_] += 1
            else:
                tag_ = tag[1:-1]
                assert tag_ in tags_
                count_[tag_] += 1
                if count_[tag_] > 0:
                    text = ''.join((text, '</', tag_, '>'))
                    count_[tag_] -= 1
        return text

    def redraw(self, widget, cr):
        if not self.__shown:
            return True
        assert self.balloon_surface is not None
        scale = self.scale
        cr.scale(scale / 100.0, scale / 100.0)
        cr.set_source_surface(self.balloon_surface, 0, 0)
        cr.set_operator(cairo.OPERATOR_SOURCE)
        cr.paint()
        cr.set_operator(cairo.OPERATOR_OVER) # restore default
        self.window.update_shape(cr.get_target())
        # draw images
        for i in range(len(self.images)):
            image_surface, (x, y) = self.images[i]
            w = image_surface.get_width()
            h = image_surface.get_height()
            if x == 'centerx':
                bw, bh = self.get_balloon_size(scaling=False)
                x = (bw - w) // 2
            else:
                try:
                    x = int(x)
                except:
                    continue
            if y == 'centery':
                bw, bh = self.get_balloon_size(scaling=False)
                y = (bh - h) // 2
            else:
                try:
                    y = int(y)
                except:
                    continue
            cr.set_source_surface(image_surface, x, y)
            cr.paint()
        # draw text
        i = self.lineno
        j = len(self.text_buffer)
        line = 0
        while line < self.lines:
            if i >= j:
                break
            x, y, w, h = self.line_regions[line]
            if self.text_buffer[i].endswith('\n[half]'):
                new_y = int(y + (self.font_height + self.line_space) // 2)
                markup = self.set_markup(i, self.text_buffer[i][:-7])
            else:
                new_y = int(y + self.font_height + self.line_space)
                markup = self.set_markup(i, self.text_buffer[i])
            self.update_line_regions(line + 1, new_y)
            self.layout.set_markup(markup, -1)
            cr.set_source_rgba(*self.text_normal_color)
            cr.move_to(x, y)
            PangoCairo.update_layout(cr, self.layout)
            PangoCairo.show_layout(cr, self.layout)
            if self.sstp_surface is not None:
                for l, c in self.sstp_marker:
                    if l == i:
                        mw = self.sstp_surface.get_width()
                        mh = self.sstp_surface.get_height()
                        self.layout.set_text(self.text_buffer[i][:c], -1)
                        text_w, text_h = self.layout.get_pixel_size()
                        mx = x + text_w
                        my = y + (self.font_height + self.line_space) // 2
                        my = my - mh // 2
                        cr.set_source_surface(self.sstp_surface, mx, my)
                        cr.paint()
            i += 1
            line += 1
        if self.side == 0 and self.sstp_message:
            self.redraw_sstp_message(widget, cr)
        if self.selection is not None:
            self.update_link_region(widget, cr, self.selection)
        self.redraw_arrow0(widget, cr)
        self.redraw_arrow1(widget, cr)
        return False

    def update_link_region(self, widget, cr, index):
        cr.save()
        sl = self.link_buffer[index][0]
        el = self.link_buffer[index][2]
        if self.lineno <= sl <= self.lineno + self.lines:
            sn = self.link_buffer[index][1]
            en = self.link_buffer[index][3]
            for n in range(sl, el + 1):
                if n - self.lineno >= len(self.line_regions):
                    break
                x, y, w, h = self.line_regions[n - self.lineno]
                if sl == el:
                    markup = self.set_markup(n, self.text_buffer[n][:sn])
                    self.layout.set_markup(markup, -1)
                    text_w, text_h =  self.layout.get_pixel_size()
                    x += text_w
                    markup = self.set_markup(n, self.text_buffer[n][sn:en])
                    self.layout.set_markup(markup, -1)
                    text_w, text_h =  self.layout.get_pixel_size()
                    w = text_w
                    start = sn
                    end = en
                elif n == sl:
                    markup = self.set_markup(n, self.text_buffer[n][:sn])
                    self.layout.set_markup(markup, -1)
                    text_w, text_h =  self.layout.get_pixel_size()
                    x += text_w
                    markup = self.set_markup(n, self.text_buffer[n][sn:])
                    self.layout.set_markup(markup, -1)
                    text_w, text_h =  self.layout.get_pixel_size()
                    w = text_w
                    start = sn
                    end = len(self.text_buffer[n])
                elif n == el:
                    markup = self.set_markup(n, self.text_buffer[n][:en])
                    self.layout.set_markup(markup, -1)
                    text_w, text_h =  self.layout.get_pixel_size()
                    w = text_w
                    start = 0
                    end = en
                else:
                    markup = self.set_markup(n, self.text_buffer[n])
                    self.layout.set_markup(markup, -1)
                    text_w, text_h =  self.layout.get_pixel_size()
                    w = text_w
                    start = 0
                    end = len(self.text_buffer[n])
                markup = self.set_markup(n, self.text_buffer[n][start:end])
                self.layout.set_markup(markup, -1)
                cr.set_source_rgba(*self.cursor_color)
                cr.rectangle(x, y, w, h)
                cr.fill()
                cr.move_to(x, y)
                cr.set_source_rgba(*self.text_active_color)
                PangoCairo.update_layout(cr, self.layout)
                PangoCairo.show_layout(cr, self.layout)
        cr.restore()

    def check_link_region(self, px, py):
        new_selection = None
        for i in range(len(self.link_buffer)):
            sl = self.link_buffer[i][0]
            el = self.link_buffer[i][2]
            if self.lineno <= sl <= self.lineno + self.lines:
                sn = self.link_buffer[i][1]
                en = self.link_buffer[i][3]
                for n in range(sl,el + 1):
                    if n - self.lineno >= len(self.line_regions):
                        break
                    x, y, w, h = self.line_regions[n - self.lineno]
                    if n == sl:
                        markup = self.set_markup(n, self.text_buffer[n][:sn])
                        self.layout.set_markup(markup, -1)
                        text_w, text_h =  self.layout.get_pixel_size()
                        x += text_w
                    if n == sl and n == el:
                        markup = self.set_markup(n, self.text_buffer[n][sn:en])
                    elif n == el:
                        markup = self.set_markup(n, self.text_buffer[n][:en])
                    else:
                        markup = self.set_markup(n, self.text_buffer[n])
                    self.layout.set_markup(markup, -1)
                    text_w, text_h =  self.layout.get_pixel_size()
                    w = text_w
                    if x <= px < x + w and y <= py < y + h:
                        new_selection = i
                        break
        if new_selection is not None:
            if self.selection != new_selection:
                sl, sn, el, en, link_id, raw_text, text = \
                    self.link_buffer[new_selection]
                self.request_parent(
                    'NOTIFY', 'notify_event',
                    'OnChoiceEnter', raw_text, link_id, self.selection)
        else:
            if self.selection is not None:
                self.request_parent('NOTIFY', 'notify_event', 'OnChoiceEnter')
        if new_selection == self.selection:
            return 0
        else:
            self.selection = new_selection
            return 1 # dirty flag

    def motion_notify(self, widget, event):
        if event.is_hint:
            ## FIXME: Use get_device_position() instead.
            _, x, y, state = widget.get_window().get_pointer()
        else:
            x, y, state = event.x, event.y, event.get_state()
        scale = self.scale
        if self.link_buffer:
            px = int(x)
            py = int(y)
            px = int(px * 100 / scale)
            py = int(py * 100 / scale)
            if self.check_link_region(px, py):
                widget.queue_draw()
        if not self.request_parent('GET', 'busy'):
            if state & Gdk.ModifierType.BUTTON1_MASK:
                if self.x_root is not None and \
                   self.y_root is not None:
                    self.dragged = True
                    x_delta = (event.x_root - self.x_root) * 100 / scale + self.x_fractions
                    y_delta = (event.y_root - self.y_root) * 100 / scale + self.y_fractions
                    self.x_fractions = x_delta - int(x_delta)
                    self.y_fractions = y_delta - int(y_delta)
                    self.request_parent(
                        'NOTIFY', 'update_balloon_offset',
                        self.side, int(x_delta), int(y_delta))
                    self.x_root = event.x_root
                    self.y_root = event.y_root
        return True

    def scroll(self, darea, event):
        px = int(event.x)
        py = int(event.y)
        scale = self.scale
        px = int(px * 100 / scale)
        py = int(py * 100 / scale)
        if event.direction == Gdk.ScrollDirection.UP:
            if self.lineno > 0:
                self.lineno = max(self.lineno - 2, 0)
                self.check_link_region(px, py)
                self.darea.queue_draw()
        elif event.direction == Gdk.ScrollDirection.DOWN:
            if self.lineno + self.lines < len(self.text_buffer):
                self.lineno = min(self.lineno + 2,
                                  len(self.text_buffer) - self.lines)
                self.check_link_region(px, py)
                self.darea.queue_draw()
        return True

    def button_press(self, darea, event):
        self.request_parent('NOTIFY', 'reset_idle_time')
        click = 1 if event.type == Gdk.EventType.BUTTON_PRESS else 2
        if self.request_parent('GET', 'is_paused'):
            self.request_parent('NOTIFY', 'notify_balloon_click',
                                event.button, click, self.side)
            return True
        # arrows
        px = int(event.x)
        py = int(event.y)
        scale = self.scale
        px = int(px * 100 / scale)
        py = int(py * 100 / scale)
        # up arrow
        w = self.arrow0_surface.get_width()
        h = self.arrow0_surface.get_height()
        x, y = self.arrow[0]
        if x <= px <= x + w and y <= py <= y + h:
            if self.lineno > 0:
                self.lineno = max(self.lineno - 2, 0)
                self.darea.queue_draw()
            return True
        # down arrow
        w = self.arrow1_surface.get_width()
        h = self.arrow1_surface.get_height()
        x, y = self.arrow[1]
        if x <= px <= x + w and y <= py <= y + h:
            if self.lineno + self.lines < len(self.text_buffer):
                self.lineno = min(self.lineno + 2,
                                  len(self.text_buffer) - self.lines)
                self.darea.queue_draw()
            return True
        # links
        if self.selection is not None:
            sl, sn, el, en, link_id, raw_text, text = \
                self.link_buffer[self.selection]
            self.request_parent('NOTIFY', 'notify_link_selection',
                                link_id, raw_text, self.selection)
            return True
        # balloon's background
        self.request_parent('NOTIFY', 'notify_balloon_click',
                            event.button, click, self.side)
        self.x_root = event.x_root
        self.y_root = event.y_root
        return True

    def button_release(self, window, event):
        x = int(event.x)
        y = int(event.y)
        scale = self.scale
        x = int(x * 100 / scale)
        y = int(y * 100 / scale)
        if self.dragged:
            self.dragged = False
        self.x_root = None
        self.y_root = None
        self.y_fractions = 0
        self.y_fractions = 0
        return True

    def clear_text(self):
        self.selection = None
        self.lineno = 0
        self.text_buffer = []
        self.meta_buffer = []
        self.link_buffer = []
        self.newline_required = 0
        self.images = []
        self.sstp_marker = []
        self.darea.queue_draw()

    def get_text_count(self):
        return self.text_count

    def reset_text_count(self):
        self.text_count = 0

    def set_newline(self):
        self.newline_required = 1

    def append_text(self, text):
        if not self.text_buffer:
            s = ''
            column = 0
            index = 0
        elif self.newline_required:
            s = ''
            column = 0
            self.newline_required = 0
            index = len(self.text_buffer)
        else:
            index = len(self.text_buffer) - 1
            s = self.text_buffer.pop(-1)
            column = len(s)
        i = len(s)
        text = ''.join((s, text))
        j = len(text)
        self.text_count += j
        p = 0
        while 1:
            if i >= j:
                self.text_buffer.append(text[p:i])
                self.draw_last_line(column)
                break
            if text[i] == '\n':
                if j >= i + 7 and text[i:i + 7] == '\n[half]':
                    self.text_buffer.append(''.join((text[p:i], '\n[half]')))
                    p = i = i + 7
                else:
                    self.text_buffer.append(text[p:i])
                    p = i = i + 1
                self.draw_last_line(column)
                column = 0
                continue
            n = i + 1
            if not self.__shown:
                self.show()
            markup = self.set_markup(index, text[p:n])
            self.layout.set_markup(markup, -1)
            text_width, text_height =  self.layout.get_pixel_size()
            if text_width > self.line_width:
                self.text_buffer.append(text[p:i])
                self.draw_last_line(column)
                column = 0
                p = i
            i = n

    def append_sstp_marker(self):
        if self.sstp_surface is None:
            return
        if not self.text_buffer:
            line = 0
            offset = 0
        else:
            line = len(self.text_buffer) - 1
            offset = len(self.text_buffer[-1])
        if self.newline_required:
            line = line + 1
            offset = 0
        self.sstp_marker.append((line, offset))
        w = self.sstp_surface.get_width()
        h = self.sstp_surface.get_height()
        i = 1
        while 1:
            space = '\u3000' * i ## FIXME
            self.layout.set_text(space, -1)
            text_w, text_h = self.layout.get_pixel_size()
            if text_w > w:
                break
            else:
                i += 1
        self.append_text(space)
        self.draw_last_line(offset)

    def append_link_in(self, link_id):
        if not self.text_buffer:
            sl = 0
            sn = 0
        else:
            sl = len(self.text_buffer) - 1
            sn = len(self.text_buffer[-1])
        self.link_buffer.append((sl, sn, sl, sn, link_id, '', ''))

    def append_link_out(self, link_id, text):
        if not text:
            return
        raw_text = text
        if not self.text_buffer:
            el = 0
            en = 0
        else:
            el = len(self.text_buffer) - 1
            en = len(self.text_buffer[-1])
        for i in range(len(self.link_buffer)):
            if self.link_buffer[i][4] == link_id:
                sl = self.link_buffer[i][0]
                sn = self.link_buffer[i][1]
                self.link_buffer.pop(i)
                self.link_buffer.insert(
                    i, (sl, sn, el, en, link_id, raw_text, text))
                break

    def append_meta(self, tag):
        if not tag:
            return
        if not self.text_buffer:
            sl = 0
            sn = 0
        else:
            sl = len(self.text_buffer) - 1
            sn = len(self.text_buffer[-1])
        self.meta_buffer.append((sl, sn, tag))

    def append_image(self, path, x, y):
        try:
            image_surface = ninix.pix.create_surface_from_file(path)
        except:
            return
        self.show()
        self.images.append((image_surface, (x, y)))
        self.darea.queue_draw()

    def draw_last_line(self, column=0):
        if not self.__shown:
            return
        line = len(self.text_buffer) - 1
        if self.lineno <= line < self.lineno + self.lines:
            x, y, w, h = self.line_regions[line - self.lineno]
            if self.text_buffer[line].endswith('\n[half]'):
                offset = line - self.lineno + 1
                new_y = int(y + (self.font_height + self.line_space) // 2)
                self.update_line_regions(offset, new_y)
            else:
                self.darea.queue_draw()
            if self.sstp_surface is not None:
                for l, c in self.sstp_marker:
                    if l == line:
                        mw = self.sstp_surface.get_width()
                        mh = self.sstp_surface.get_height()
                        self.layout.set_text(self.text_buffer[l][:c], -1)
                        text_w, text_h = self.layout.get_pixel_size()
                        mx = x + text_w
                        my = y + (self.font_height + self.line_space) // 2
                        my = my - mh // 2
                        cr = self.darea.get_window().cairo_create()
                        cr.set_source_surface(self.sstp_surface, mx, my)
                        cr.paint()
                        del cr
        else:
            self.darea.queue_draw()
            if self.autoscroll:
                while line >= self.lineno + self.lines:
                    self.lineno += 1
                    self.darea.queue_draw()


class CommunicateWindow(metaclass=abc.ABCMeta):

    NAME = ''
    ENTRY = ''

    def __init__(self):
        self.request_parent = lambda *a: None # dummy
        self.window = None

    def set_responsible(self, request_method):
        self.request_parent = request_method

    def new(self, desc, balloon):
        if self.window:
            self.window.destroy()
        self.window = Gtk.Window()
        screen = self.window.get_screen()
        if self.window.is_composited():
            visual = screen.get_rgba_visual()
        else:
            visual = screen.get_system_visual()
            logging.debug('screen does NOT support alpha.\n')
        assert visual is not None
        self.window.set_visual(visual)
        self.window.set_title('communicate')
        self.window.set_decorated(False)
        self.window.set_resizable(False)
        self.window.connect('delete_event', self.delete)
        self.window.connect('key_press_event', self.key_press)
        self.window.connect('button_press_event', self.button_press)
        self.window.connect('drag_data_received', self.drag_data_received)
        # DnD data types
        dnd_targets = [Gtk.TargetEntry.new('text/plain', 0, 0)]
        self.window.drag_dest_set(Gtk.DestDefaults.ALL, dnd_targets,
                                  Gdk.DragAction.COPY)
        self.window.drag_dest_add_text_targets()
        self.window.set_events(Gdk.EventMask.BUTTON_PRESS_MASK)
        self.window.set_modal(True)
        self.window.set_position(Gtk.WindowPosition.CENTER)
        self.window.realize()
        w = desc.get_with_type('communicatebox.width', int, 250)
        h = desc.get_with_type('communicatebox.height', int, -1)
        self.entry = Gtk.Entry()
        self.entry.connect('activate', self.activate)
        self.entry.set_inner_border(None)
        self.entry.set_has_frame(False)
        font_desc = Pango.FontDescription()
        font_desc.set_size(9 * 3 / 4 * Pango.SCALE) # XXX
        self.entry.modify_font(font_desc)
        self.entry.set_size_request(w, h)
        self.entry.show()
        surface = None
        if balloon:
            path, config = balloon
            # load pixbuf
            try:
                surface = ninix.pix.create_surface_from_file(path)
            except:
                surface = None
        if surface is not None:
            darea = Gtk.DrawingArea()
            darea.set_events(Gdk.EventMask.EXPOSURE_MASK)
            darea.connect('draw', self.redraw, surface)
            darea.show()
            x = desc.get_with_type('communicatebox.x', int, 10)
            y = desc.get_with_type('communicatebox.y', int, 20)
            overlay = Gtk.Overlay()
            self.entry.set_margin_left(x)
            self.entry.set_margin_top(y)
            self.entry.set_halign(Gtk.Align.START)
            self.entry.set_valign(Gtk.Align.START)
            overlay.add_overlay(self.entry)
            overlay.add(darea)
            overlay.show()
            self.window.add(overlay)
            w = surface.get_width()
            h = surface.get_height()
            darea.set_size_request(w, h)
        else:
            box = Gtk.HBox(spacing=10)
            box.set_border_width(10)
            if self.ENTRY:
                label = Gtk.Label(label=self.ENTRY)
                box.pack_start(label, False, True, 0)
                label.show()
            box.pack_start(self.entry, True, True, 0)
            self.window.add(box)
            box.show()

    def drag_data_received(self, widget, context, x, y, data, info, time):
        self.entry.set_text(data.get_text())

    def redraw(self, widget, cr, surface):
        cr.set_source_surface(surface, 0, 0)
        cr.set_operator(cairo.OPERATOR_SOURCE)
        cr.paint()
        region = Gdk.cairo_region_create_from_surface(cr.get_target())
        # XXX: to avoid losing focus in the text input region
        x = self.entry.get_margin_left()
        y = self.entry.get_margin_top()
        w = self.entry.get_allocated_width()
        h = self.entry.get_allocated_height()
        region.union(cairo.RectangleInt(x, y, w, h))
        self.window.input_shape_combine_region(region)

    def destroy(self):
        if self.window:
            self.window.destroy()
            self.window = None

    def delete(self, widget, event):
        self.window.hide()
        self.cancel()
        return True

    def key_press(self, widget, event):
        if event.keyval == Gdk.KEY_Escape:
            self.window.hide()
            self.cancel()
            return True
        return False

    def button_press(self, widget, event):
        if event.button in [1, 2]:
            self.window.begin_move_drag(
                event.button, int(event.x_root), int(event.y_root),
                Gtk.get_current_event_time())
        return True

    def activate(self, widget):
        self.window.hide()
        self.enter()
        return True

    def show(self, default=''):
        self.entry.set_text(default)
        self.window.show()

    @abc.abstractmethod
    def enter(self):
        pass

    @abc.abstractmethod
    def cancel(self):
        pass


class CommunicateBox(CommunicateWindow):

    NAME = 'communicatebox'
    ENTRY = 'Communicate'

    def new(self, desc, balloon):
        CommunicateWindow.new(self, desc, balloon)
        self.window.set_modal(False)

    def delete(self, widget, event):
        self.window.hide()
        self.cancel()
        self.request_parent('NOTIFY', 'reset_user_interaction')
        return True

    def key_press(self, widget, event):
        if event.keyval == Gdk.KEY_Escape:
            self.window.hide()
            self.cancel()
            self.request_parent('NOTIFY', 'reset_user_interaction')
            return True
        return False

    def activate(self, widget):
        self.enter()
        self.entry.set_text('')
        return True

    def enter(self):
        self.send(self.entry.get_text())

    def cancel(self):
        self.request_parent('NOTIFY', 'notify_event',
                            'OnCommunicateInputCancel', '', 'cancel')

    def send(self, data):
        if data is not None:
            self.request_parent('NOTIFY', 'notify_event',
                                'OnCommunicate', 'user', data)


class TeachBox(CommunicateWindow):

    NAME = 'teachbox'
    ENTRY = 'Teach'

    def enter(self):
        self.send(self.entry.get_text())

    def cancel(self):
        self.request_parent('NOTIFY', 'notify_event',
                            'OnTeachInputCancel', '', 'cancel')
        self.request_parent('NOTIFY', 'reset_user_interaction')

    def send(self, data):
        self.request_parent('NOTIFY', 'notify_user_teach', data)
        self.request_parent('NOTIFY', 'reset_user_interaction')


class InputBox(CommunicateWindow):

    NAME = 'inputbox'
    ENTRY = 'Input'

    def new(self, desc, balloon):
        CommunicateWindow.new(self, desc, balloon)
        self.symbol = None
        self.limittime = -1

    def set_symbol(self, symbol):
        self.symbol = symbol
        
    def set_limittime(self, limittime):
        try:
            limittime = int(limittime)
        except ValueError:
            limittime = -1
        self.limittime = limittime

    def show(self, default):
        if default is not None:
            try:
                text = str(default)
            except:
                text = ''
        else:
            text = ''
        if int(self.limittime) < 0:
            self.timeout_id = None
        else:
            self.timeout_id = GLib.timeout_add(self.limittime, self.timeout)
        CommunicateWindow.show(self, text)

    def timeout(self):
        self.window.hide()
        self.send('timeout', timeout=True)

    def enter(self):
        self.send(self.entry.get_text())

    def cancel(self):
        self.send(None, cancel=True)

    def close(self, symbol):
        if self.symbol is None:
            return
        if symbol != '__SYSTEM_ALL_INPUT__' and self.symbol != symbol:
            return
        self.window.hide()
        self.cancel()

    def send(self, data, cancel=False, timeout=False):
        if self.timeout_id is not None:
            GLib.source_remove(self.timeout_id)            
        if data is None:
            data = ''
        ## CHECK: symbol
        if cancel:
            self.request_parent('NOTIFY', 'notify_event',
                                'OnUserInputCancel', '', 'cancel')
        elif timeout and \
                self.request_parent('GET', 'notify_event',
                                    'OnUserInputCancel', '', 'timeout'):
            pass
        elif self.symbol == 'OnUserInput' and \
           self.request_parent('GET', 'notify_event', 'OnUserInput', data):
            pass
        elif self.request_parent('GET', 'notify_event',
                                 'OnUserInput', self.symbol, data):
            pass
        elif self.request_parent('GET', 'notify_event', self.symbol, data):
            pass
        self.symbol = None
        self.request_parent('NOTIFY', 'reset_user_interaction')


class PasswordInputBox(InputBox):

    NAME = 'passwordinputbox'
    ENTRY = 'PasswordInput'

    def new(self, desc, balloon):
        InputBox.new(self, desc, balloon)
        self.entry.set_visibility(False)


def test():
    pass

if __name__ == '__main__':
    test()
